/*
 * Copyright (c) 1997, 1998, 1999
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD$
 */

/*
 * This is a PCI shim for the Aironet PC4500/4800 wireless network
 * driver. Aironet makes PCMCIA, ISA and PCI versions of these devices,
 * which all have basically the same interface. The ISA and PCI cards
 * are actually bridge adapters with PCMCIA cards inserted into them,
 * however they appear as normal PCI or ISA devices to the host.
 *
 * The guts of the driver is in sys/i386/isa/{if_an.c|if_anreg.h}.
 * All we do here is handle the PCI probe and attach and set up an
 * interrupt handler entry point. The PCI version of the card uses
 * a PLX 9050 PCI to "dumb bus" bridge chip, which provides us with
 * multiple PCI address space mappings. The primary mapping at PCI
 * register 0x14 is for the PLX chip itself, *NOT* the Aironet card.
 * The I/O address of the Aironet is actually at register 0x18, which
 * is the local bus mapping register for bus space 0. There are also
 * registers for additional register spaces at registers 0x1C and
 * 0x20, but these are unused in the Aironet devices. To find out
 * more, you need a datasheet for the 9050 from PLX, but you have
 * to go through their sales office to get it. Bleh.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>

#include <machine/bus_pio.h>
#include <machine/bus.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#include "an.h"

#ifndef lint
static const char rcsid[] =
 "$FreeBSD$";
#endif

#if NAN > 0

#include <i386/isa/if_anreg.h>

struct an_type {
	u_int16_t		an_vid;
	u_int16_t		an_did;
	char			*an_name;
};

/* These are exported from src/i386/isa/if_an.c. */
extern void an_intr			__P((void *));
extern int an_attach			__P((struct an_softc *));
extern unsigned long an_count_pnp;

#define AIRONET_VENDORID	0x14B9
#define AIRONET_DEVICEID_4500	0x4500
#define AIRONET_DEVICEID_4800	0x4800
#define AIRONET_DEVICEID_4xxx	0x0001
#define AN_PCI_PLX_LOIO		0x14	/* PLX chip iobase */
#define AN_PCI_LOIO		0x18	/* Aironet iobase */

static struct an_type an_devs[] = {
	{ AIRONET_VENDORID, AIRONET_DEVICEID_4500, "Aironet PCI4500" },
	{ AIRONET_VENDORID, AIRONET_DEVICEID_4800, "Aironet PCI4800" },
	{ AIRONET_VENDORID, AIRONET_DEVICEID_4xxx, "Aironet PCI4500/4800" },
	{ 0, 0, NULL }
};

unsigned long an_count = NAN;

static const char *an_probe_pci	__P((pcici_t, pcidi_t));
static void an_attach_pci	__P((pcici_t, int));
static void an_intr_pci		__P((void *));

static const char *an_probe_pci(config_id, device_id)
	pcici_t			config_id;
	pcidi_t			device_id;
{
	struct an_type		*t;

	t = an_devs;

	while(t->an_name != NULL) {
		if ((device_id & 0xFFFF) == t->an_vid &&
		    ((device_id >> 16) & 0xFFFF) == t->an_did) {
			an_count_pnp++;
			return(t->an_name);
		}
		t++;
	}

	return(NULL);
}

static void an_attach_pci(config_id, unit)
	pcici_t			config_id;
	int			unit;
{
	int			s;
	u_int32_t		command;
	struct an_softc		*sc;

	s = splimp();

	sc = malloc(sizeof(struct an_softc), M_DEVBUF, M_NOWAIT);
	if (sc == NULL) {
		printf("an%d: no memory for softc struct!\n", unit);
		goto fail;
	}

	bzero(sc, sizeof(struct an_softc));

	/*
	 * Map control/status registers.
 	 */
	command = pci_conf_read(config_id, PCI_COMMAND_STATUS_REG);
	command |= PCIM_CMD_PORTEN;
	pci_conf_write(config_id, PCI_COMMAND_STATUS_REG, command);
	command = pci_conf_read(config_id, PCI_COMMAND_STATUS_REG);

	if (!(command & PCIM_CMD_PORTEN)) {
		printf("an%d: failed to enable I/O ports!\n", unit);
		free(sc, M_DEVBUF);
		goto fail;
	}

	/* Map the PLX chip. */
	if (!pci_map_port(config_id, AN_PCI_PLX_LOIO,
	    (pci_port_t *)&(sc->an_bhandle_p))) {
		printf("an%d: couldn't map ports\n", unit);
		free(sc, M_DEVBUF);
		goto fail;
	}

	/* Now map the Aironet registers. */
	if (!pci_map_port(config_id, AN_PCI_LOIO,
	    (pci_port_t *)&(sc->an_bhandle))) {
		printf("an%d: couldn't map ports\n", unit);
		free(sc, M_DEVBUF);
		goto fail;
	}
	sc->an_btag = I386_BUS_SPACE_IO;

	/* Allocate interrupt */
	if (!pci_map_int(config_id, an_intr_pci, sc, &net_imask)) {
		printf("an%d: couldn't map interrupt\n", unit);
		free(sc, M_DEVBUF);
		goto fail;
	}

	sc->an_unit = unit;

	if (an_attach(sc)) {
		printf("an%d: attach failed\n", unit);
		free(sc, M_DEVBUF);
	};

fail:
	splx(s);

	return;
}

static void an_intr_pci(xsc)
	void			*xsc;
{
	an_intr(xsc);

	return;
}

static struct pci_device an_device = {
	"an",
	an_probe_pci,
	an_attach_pci,
	&an_count,
	NULL
};
#ifdef COMPAT_PCI_DRIVER
COMPAT_PCI_DRIVER(an, an_device);
#else
DATA_SET(pcidevice_set, an_device);
#endif

#endif
